# Models.pm 
#
# Extracts features that constitute good clues for the disambiguation 
# of a target word -- e.g. noun, verb, noun with a certain property, etc.
#
# Copyright (c) 2004 Rada Mihalcea
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#
# CHANGE LOG
#
# Date          Author          Change
# =======================================================================
# 11-01-2004    Rada            Create Models.pm 
# =======================================================================


#=======================================================#
#                                                       #
#               CONTEXT-BASED MODELS                    #
#                                                       #
#=======================================================#


# =======================================================================
# modelNN1
#
# Extracts features that constitute good clues for the disambiguation 
# of a target noun. This is learning model 1 for nouns. It currently
# extracts the first noun/verb before the target noun.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 11-01-2004    Rada            Make this a separate subroutine
# 11-20-2003    Rada            Initial version
# ========================================================================

sub modelNN1 {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /NN/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print information for the first word to the left
		# look only for adj, noun, verb
		my $found = 0;
		for(my $k = $index - 1; 
		    $k >= 0 && $index - $k <= $wordsBefore; 
		    $k--) {
		    if($sentence[$k]->{pos} =~ /NN/ 
		       || $sentence[$k]->{pos} =~ /VB/) {
			print $FILE "$sentence[$k]->{lemma} ";
			print $FILE "$sentence[$k]->{pos} ";
			$found = 1;
			last;
		    }
		}
		if(0 == $found) {
		    print $FILE "? ";
		    print $FILE "? ";
		}
		
		    
		# print information for the first word to the right
		# look only for noun, verb
		$found = 0;
		for($k = $index + 1; 
		    $k < $lenSentence && $k - $index <= $wordsAfter; 
		    $k++) {
		    if($sentence[$k]->{pos} =~ /NN/
		       || $sentence[$k]->{pos} =~ /VB/) {
			# print $FILE "$sentence[$k]->{lemma} ";
			# print $FILE "$sentence[$k]->{pos} ";
			$found = 1;
			last;
		    }
		}
		if(0 == $found) {
		    # print $FILE "? ";
		    # print $FILE "? ";
		}
		
		
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}



# =======================================================================
# modelJJ1
#
# Extracts features that constitute good clues for the disambiguation 
# of a target adjective. This is learning model 1 for adjectives -- it 
# is based on word dependencies. It currently extracts the first noun 
# after the target adjective. 
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
# ========================================================================
# Date          Author          Change
# 11-01-2004    Rada            Make this a separate subroutine
# 11-20-2003    Rada            Initial version
# ========================================================================


sub modelJJ1 {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /JJ/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print information for the first word to the right
		# look only for nouns
		$found = 0;
		for(my $k = $index + 1;
		    $k < $lenSentence && $k - $index <= $wordsAfter;
		    $k++) {
		    if($sentence[$k]->{pos} =~ /NN/) {
			print $FILE "$sentence[$k]->{lemma} ";
			print $FILE "$sentence[$k]->{pos} ";
			$found = 1;
			last;
		    }
		}		
		if(0 == $found) {
		    print $FILE "? ";
		    print $FILE "? ";
		}
		
		    

		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}


# =======================================================================
# modelJJ2
#
# Extracts features that constitute good clues for the disambiguation 
# of a target adjective. This is learning model 2 for adjectives -- it 
# is context based. It currently extracts surrounding context of the target 
# adjective
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
# ========================================================================
# Date          Author          Change
# 11-01-2004    Rada            Make this a separate subroutine
# 11-20-2003    Rada            Initial version
# ========================================================================

sub modelJJ2 {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /JJ/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print information about the context
		# print information for the first word to the left and right
		for(my $k = $index - 1;
		    $k <= $index + 1;
		    $k++) {
		    if($k == $index) {
			next;
		    }
		    if($k >= 0 && $k < $lenSentence) {
			print $FILE "$sentence[$k]->{lemma} ";
			print $FILE "$sentence[$k]->{pos} ";
		    }
		    else {
			print $FILE "? ";
			print $FILE "? ";
		    }
		}

		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}



# =======================================================================
# modelVB1
#
# Extracts features that constitute good clues for the disambiguation 
# of a target verb. This is learning model 1 for verbs. It is context-based.
# It currently extracts the immediate surrounding context for the target verb.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
# ========================================================================
# Date          Author          Change
# 11-01-2004    Rada            Make this a separate subroutine
# 11-20-2003    Rada            Initial version
# ========================================================================


sub modelVB1 {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /VB/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print information about the context
		# print information for the first word to the left and right
		    for(my $k = $index - 1;
                        $k <= $index + 1;
                        $k++) {
			if($k == $index) {
			    next;
			}
                        if($k >= 0 && $k < $lenSentence) {
                            print $FILE "$sentence[$k]->{lemma} ";
                            print $FILE "$sentence[$k]->{pos} ";
                        }
                        else {
                            print $FILE "? ";
                            print $FILE "? ";
                        }
                    }
		
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}


#=======================================================#
#                                                       #
#               COLLOCATION-BASED MODELS                #
#                                                       #
#=======================================================#

 
# =======================================================================
# modelNNCollocations
#
# Extracts features that constitute good clues for the disambiguation 
# of a target noun. This learning model is based on collocations surrounding
# the target noun.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 11-05-2004    Rada            Initial version
# ========================================================================

sub modelNNCollocations {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /NN/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print collocation information
		# [0-1 words before] X [0-2 words after]
		for(my $before = 0; $before <= 1; $before++) {
		    for(my $after = 0; $after <= 1; $after++) {  
			my $collocation = "";
			for(my $k = $index - $before; $k <= $index + $after;
			    $k++) {
			    if($k < 0 || $k >= $lenSentence) {
				$collocation .= "_?";
			    }
			    else {
				$collocation .= "_".$sentence[$k]->{word};
			    }
			}
			print $FILE "$collocation ";
		    }
		}
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}

     
# =======================================================================
# modelJJCollocations
#
# Extracts features that constitute good clues for the disambiguation 
# of a target adjective. This learning model is based on collocations 
# surrounding the target adjective.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 11-05-2004    Rada            Initial version
# ========================================================================

sub modelJJCollocations {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /JJ/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print collocation information
		# [0-1 words before] X [0-2 words after]
		for(my $before = 0; $before <= 1; $before++) {
		    for(my $after = 0; $after <= 1; $after++) {  
			my $collocation = "";
			for(my $k = $index - $before; $k <= $index + $after;
			    $k++) {
			    if($k < 0 || $k >= $lenSentence) {
				$collocation .= "_?";
			    }
			    else {
				$collocation .= "_".$sentence[$k]->{word};
			    }
			}
			print $FILE "$collocation ";
		    }
		}
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}

     
# =======================================================================
# modelVBCollocations
#
# Extracts features that constitute good clues for the disambiguation 
# of a target verb. This learning model is based on collocations 
# surrounding the target verb.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 11-05-2004    Rada            Initial version
# ========================================================================

sub modelVBCollocations {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {
	    if($sentence[$index]->{pos} =~ /VB/ &&
	       $sentence[$index]->{lemma} ne "?") {
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    
		# print collocation information
		# [0-1 words before] X [0-2 words after]
		for(my $before = 0; $before <= 1; $before++) {
		    for(my $after = 0; $after <= 1; $after++) {  
			my $collocation = "";
			for(my $k = $index - $before; $k <= $index + $after;
			    $k++) {
			    if($k < 0 || $k >= $lenSentence) {
				$collocation .= "_?";
			    }
			    else {
				$collocation .= "_".$sentence[$k]->{word};
			    }
			}
			print $FILE "$collocation ";
		    }
		}
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}


#=======================================================#
#                                                       #
#             TEMPLATE FOR A NEW MODEL SUBROUTINE       #
#                                                       #
#=======================================================#


# =======================================================================
# subroutineTemplate
#
# Extracts features that constitute good clues for the disambiguation 
# of a target word. Note that this subroutine *cannot* be used as is
# Rather, use this subroutine as a starting point to implement your
# own models.
#
# PARAMETERS
# sentenceRef            reference to sentence containing instances of 
#                        type Word
# FILE                   file handle for the file where features vectors
#                        will be stored
# wordsBefore/wordsAfter define the window for feature extraction
#
# 
# CHANGE LOG
# ========================================================================
# Date          Author          Change
# 11-13-2004    Rada            Create the template
# ========================================================================


sub subroutineTemplate {
    my ($sentenceRef, $FILE, $wordsBefore, $wordsAfter) = @_;

    my @sentence = @$sentenceRef;
    my $lenSentence = scalar(@sentence);
    
    for(my $index = 0; $index < $lenSentence; $index++) {
	if($sentence[$index]->{type} eq "wsd") {

	    # CHANGE THE FOLLOWING IF STATEMENT
	    # The following "if" statement defines a constraint
	    # on the target word. The model implemented in this
            # subroutine will only address the words for which
	    # this "if" statement evaluates to true.

	    if($sentence[$index]->{pos} =~ /JJ/ &&                    #CHANGE
	       $sentence[$index]->{lemma} ne "?") {                   #CHANGE
		
		# print the id
		print $FILE "$sentence[$index]->{id} ";
		
		# print the word itself and the POS
		print $FILE "$sentence[$index]->{lemma} ";
		print $FILE "$sentence[$index]->{pos} ";
		    

		# CHANGE THE FOLLOWING STATEMENTS TO ADD YOUR
		# OWN FEATURE EXTRACTOR
		$found = 0;
		for(my $k = $index + 1;                               #CHANGE
		    $k < $lenSentence && $k - $index <= $wordsAfter;  #CHANGE
		    $k++) {                                           #CHANGE
		    if($sentence[$k]->{pos} =~ /NN/) {                #CHANGE
			print $FILE "$sentence[$k]->{lemma} ";        #CHANGE
			print $FILE "$sentence[$k]->{pos} ";          #CHANGE
			$found = 1;                                   #CHANGE
			last;                                         #CHANGE
		    }                                                 #CHANGE
		}		                                      #CHANGE
		if(0 == $found) {                                     #CHANGE
		    print $FILE "? ";                                 #CHANGE
		    print $FILE "? ";                                 #CHANGE
		}                                                     #CHANGE
		
		    
		
		# and print the "class" -- the word with the corresponding sense
		print $FILE "$sentence[$index]->{lemma}#$sentence[$index]->{wnsn}\n";
	    }
	}
    }
}


   

1;
