#!/usr/bin/perl -w

# senseLearnerWSD.pl
#
# Disambiguate all words in open text, based on information learned
# from sense tagged data (e.g. SemCor)
# 
# Copyright (c) 2004 Rada Mihalcea
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#
#
# INPUT:    file formatted using the SemCor XML format
# OUTPUT:   file annotated with word senses, also in SemCor XML format
#           [optionally evaluation figures]
#
# CHANGE LOG
#
# Date          Author          Change
# =======================================================================
# 11-01-2004    Rada            add modules NNFeatures, VBFeatures, JJFeatures 
#                               for additional flexibility
# 01-20-2003    Rada            initial version
# =======================================================================


use Getopt::Long;
use Word;
use Models;


# set default options and read command line options
$WINDOW_SIZE = 3;
$MODELS_DIR = "models";
if(scalar(@ARGV) == 0) {
    &displayHelp;
    exit;
}

&getCommandLineOptions(\$wordsBefore, \$wordsAfter, \$inFile, \$outFile, \$evaluate, \@models);


# check if training files exist  
&checkTrainingFiles(\@models);

# read in text and store it in the @discourse strucure
my (@discourse);
&readinText(\@discourse, $inFile);


# apply SenseLearner algorithm
&applySenseLearner(\@discourse, \@models);


if(defined($evaluate)) {
# evaluate the disambiguation
    &evaluate(\@discourse);
}


# store the disambiguation results in output file
&printResults(\@discourse, $outFile);


# remove all intermediate files
`rm tmp.test.*`;





########################################################################
#
#  S U B R O U T I N E S
#
########################################################################


# ======================================================================
# checkTrainingFiles
#
# checks on the existence of training files
# 
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-01-2004    Rada            Initial version
#=======================================================================
sub checkTrainingFiles {
    my($modelsRef) = @_;

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);
    
    for(my $i=0; $i < $lenModels; $i++) {
	(-e "$MODELS_DIR/training.".$models[$i]) || die "Could not open training file $MODELS_DIR/training.".$models[$i];
    }
}



# ======================================================================
# applySenseLearner
#
# apply the sense learner algorithm
# - build feature vectors for all adj, nouns, verbs
# and do a learning using the given training files
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-01-2004    Rada            Add calls to extractFeatures functions
# 03-22-2004    Rada            Initial version
#=======================================================================
sub applySenseLearner {
    my ($discourseRef, $modelsRef) = @_;

    my @discourse = @$discourseRef;
    my $lenDiscourse = scalar(@discourse);

    my @models = @$modelsRef;
    my $lenModels = scalar(@models);
    

    # EXTRACT FEATURES, ACCORDING TO EACH MODEL
    # and store them in tmp.test.xx files
    
    my @fileHandles;
    
    # open files to store feature vectors
    for(my $m=0; $m < $lenModels; $m++) {
	open $fileHandles[$m], ">tmp.test.$models[$m]";
	(-w $fileHandles[$m]) || die "Could not open test file tmp.test.$models[$m]\n";
    } 
    
    
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};
    
	# extract features for all models for this sentence
	for(my $m=0; $m < $lenModels; $m++) {
	    &{$models[$m]}(\@sentence, $fileHandles[$m], $wordsBefore, $wordsAfter);
	}
	
    }

    # close files 
    for(my $m=0; $m < $lenModels; $m++) {
	close $fileHandles[$m];
    } 

    

    # LEARNING STEP
    # Invoke the MBL learner through system calls
    for(my $m=0; $m < $lenModels; $m++) {
	`Timbl -f $MODELS_DIR/training.$models[$m] -t tmp.test.$models[$m] -o tmp.test.$models[$m].mbl`;
    } 

    # And collect the predictions made by the learner into a single file
    # Make sure the concept predicted is the same as the target word
    my $awkcommand = "awk '{print \$1\" \"\$2\" \"\$NF}' ";
    for(my $m=0; $m < $lenModels; $m++) {
	$awkcommand .= "tmp.test.$models[$m].mbl ";
    }
    $awkcommand .= " | sed 's/[\,\#]/ /g' | awk '{if (\$2 == \$3) print \$1\" \"\$4}' | sort | uniq > tmp.test.results.mbl";
    `$awkcommand` ;
    
    
    # Read the predicted senses
    open PREDICTIONS, "<tmp.test.results.mbl";
    (-r PREDICTIONS) || die "Could not open result file with sense predictions\n";
    
    while($line = <PREDICTIONS>) {
	chomp $line;
	my ($id, $sense) = split /\s+/, $line;
	
	# make sure only *one* label is provided for an instance
	# if several labels are available, it means that various models
	# did not agree on the classification, and thus no prediction 
	# is made
	if(exists($predictions{$id})) {
	    delete($predictions{$id});
	}
	
	else {
	    $predictions{$id} = $sense;
	}
    }
    close PREDICTIONS;

    
    # APPLY THE PREDICTIONS ON THE TEST FILE
    
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sentence = @{$discourse[$d]};
	
	my $lenSentence = scalar(@sentence);
	
	for(my $i=0; $i < $lenSentence; $i++) {
	    if($sentence[$i]->{type} eq "wsd") {
		if(exists $predictions{$sentence[$i]->{id}}) {
		    $sentence[$i]->{predicted} = $predictions{$sentence[$i]->{id}};
		}
	    }
	}
	@{$discourse[$d]} = @sentence;
    }
    
    @$discourseRef = @discourse;
    
    
}
 


# ======================================================================
# readinText
#
# read all sentences
# store Word instances for each word, punctuation
# the resulting structure (@discourse) is a matrix
#  - sentences are on rows
#  - for each sentence (row) words are on columns
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-03-2004    Rada            initial version
#=======================================================================

# a global variable incremented for each instance -- represents
# a unique id to be assigned to those instances that do not have
# an id in the input file
$wordId = 0;

sub readinText {
    my ($discourseRef, $inFile) = @_;
    my @discourse = @$discourseRef;
  
    my $sentIndex = 0;
    my $wordIndex = 0;


    # load input data
    # open the SemCor-like file, and start reading it
    open INFILE, "<$inFile";
    (-r INFILE) || die "Could not open input file $inFile. Please check path and make sure is valid\n";

    while(my $line = <INFILE>) {
	chomp $line;
	$line =~ s/^\s+//;
	$line =~ s/\s+$//;
	
	# if end of sentence, reset wordIndex, and increase sentIndex
	if($line =~ /^\<\/s\>/) {
	    $wordIndex=0;
	    $sentIndex++;
	    next;
	}    
	
	# if inside sentence, store Word-s
	if($line =~ /^\<wf/ ||
	   $line =~ /^\<punc/) {
	    $word = new Word($line);
	    if($word->{id} eq "?") {
		$wordId++;
		$word->{id} = "AUTO.".$wordId;
	    }
	    if(defined $word) {
		$discourse[$sentIndex][$wordIndex++] = $word;
	    }
	}
    }
    
    @$discourseRef = @discourse;
}


#=======================================================================
# printResults
#
# print the results of the disambiguation program, in a Semcor-like
# format
# 
# Note! It also prints out the lexsn field -- but this slows 
# down the program quite a bit, since it loads index.sense in memory
# Remove the section for adding lexsn if not needed
#
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 03-03-2004    Rada            initial version
#=======================================================================

sub printResults {
    my ($discourseRef, $outFile) = @_;
    my @discourse = @$discourseRef;
  
    #prepare the output file for writing
    open OUTFILE, ">$outFile";
    (-w OUTFILE) || die "Could not open output file for writing $outFile.\n Please check path and make sure is valid\n";
 
    my $lenDiscourse = scalar(@discourse);
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sent = @{$discourse[$d]};
	

	if($d != 0) {
	    print OUTFILE "</s>\n";
	}
	print OUTFILE "<s snum=", $d+1, ">\n";
	
	my $lenSentence = scalar(@sent);
	for(my $i=0; $i < $lenSentence; $i++) {
	    
	    # open class word
	    if($sent[$i]->{type} eq "wsd") {
		print OUTFILE "<wf cmd=done ";
		if($sent[$i]->{id} !~ "AUTO") {
		    print OUTFILE "id=", $sent[$i]->{id}," ";
		}
		print OUTFILE "pos=",$sent[$i]->{pos}," ";
		print OUTFILE "lemma=",$sent[$i]->{lemma}," ";
		print OUTFILE "wnsn=",$sent[$i]->{wnsn}," ";
		print OUTFILE "predicted=",$sent[$i]->{predicted};
		
		print OUTFILE ">",$sent[$i]->{word}, "</wf>\n";
	    }
	    
	    elsif ($sent[$i]->{type} eq "punct") {
		print OUTFILE "<punc>",$sent[$i]->{word},"</punc>\n";
	    }
	    else {
		print OUTFILE "<wf cmd=ignore ";
		print OUTFILE "pos=",$sent[$i]->{pos}," ";
		print OUTFILE "lemma=",lc($sent[$i]->{word});
		print OUTFILE ">",$sent[$i]->{word}, "</wf>\n";
	    }
	}

    }
	    
    
    print OUTFILE "</s>\n";
    close OUTFILE;
		
}
   



#=======================================================================
# evaluate
#
# evaluates the predicted senses versus the correctly assigned 
# senses
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 02-27-2004    Rada            initial version
#=======================================================================

sub evaluate {
    my ($discourseRef) = @_;

    my @discourse = @$discourseRef;

    my $totalWords = 0;
    my $correctWords = 0;

    my $totalIdWords = 0;
    my $correctIdWords = 0;

    my $totalWordsAddressed = 0;
    my $totalIdWordsAddressed = 0;
    my $correctWordsAddressed = 0;
    my $correctIdWordsAddressed = 0;


    my $lenDiscourse = scalar(@discourse);
    for(my $d=0; $d<$lenDiscourse; $d++) {
	my @sent = @{$discourse[$d]};
	
	my $lenSentence = scalar(@sent);

	for(my $i=0; $i < $lenSentence; $i++) {
	    if($sent[$i]->{type} eq "wsd" &&
	       $sent[$i]->{wnsn} != 0) {  # do not count words that cannot be evaluated
		
		# total words
		$totalWords++;
		if($sent[$i]->{predicted} > 0) { 
		    $totalWordsAddressed++;
		}
		if($sent[$i]->{id} !~ "AUTO") {
		    $totalIdWords++;
		    if($sent[$i]->{predicted} > 0) {
			$totalIdWordsAddressed++;
		    }
		}
		
		# correctly tagged words -- words 
		# with no sense predicted (0) are considered
		# as tagged with sense #1
		if(($sent[$i]->{predicted} > 0 &&
		    $sent[$i]->{wnsn} == $sent[$i]->{predicted}) ||
		   ($sent[$i]->{predicted} <= 0 &&
		    $sent[$i]->{wnsn} == 1)) {
		    $correctWords++;
		}
		if($sent[$i]->{predicted} > 0 &&
		   $sent[$i]->{wnsn} == $sent[$i]->{predicted}) {
		    $correctWordsAddressed++;
		}

		if($sent[$i]->{id} !~ "AUTO") {
		    if(($sent[$i]->{predicted} > 0 &&
			$sent[$i]->{wnsn} == $sent[$i]->{predicted}) ||
		       ($sent[$i]->{predicted} <= 0 &&
			$sent[$i]->{wnsn} == 1)) {
			$correctIdWords++;
		    }
		    if($sent[$i]->{predicted} > 0 &&
		       $sent[$i]->{wnsn} == $sent[$i]->{predicted}) {
			$correctIdWordsAddressed++;
		    }
		}
		
	    }
	}
    }
    
    print "Evaluation:\n";
    print "==========\n\n";
    print "ALL OPEN CLASS WORDS:\n\n";

    print "[partial coverage, only words addressed by model(s)]\n";
    if(0 != $totalWordsAddressed) {
	print "Total words addressed: $totalWordsAddressed\n";
	print "Correct words addressed: $correctWordsAddressed (",
	$correctWordsAddressed/$totalWordsAddressed,")\n\n";
    }
    
    print "[complete coverage, using a back-off to first sense]\n";    
    if(0 != $totalWords) {
	print "Total words: $totalWords\n";
	print "Correct words: $correctWords (",
	$correctWords/$totalWords,")\n";
    }

    print "\n\n=========\n\n";
    print "ALL WORDS MARKED WITH AN ID (words with an id=xxx field):\n\n";
    print "[partial coverage, only words addressed by model(s)]\n";
    if(0 != $totalIdWordsAddressed) {
	print "Total IDwords addressed: $totalIdWordsAddressed\n";
	print "Correct IDwords addressed: $correctIdWordsAddressed (",
	$correctIdWordsAddressed/$totalIdWordsAddressed,")\n\n";
    }

    print "[complete coverage, using a back-off to first sense]\n"; 
    if(0 != $totalIdWords) {
	print "Total IDwords: $totalIdWords\n";
	print "Correct IDwords: $correctIdWords (",
	$correctIdWords/$totalIdWords,")\n\n\n";
    }

}


#=======================================================================
# getCommandLineOptions                                       
#
# Get the command lines options
# Check existence/validity of mandatory options
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-24-2003    Rada            initial version
#=======================================================================

sub getCommandLineOptions {
    my ($wordsBeforeRef,$wordsAfterRef, $sourceFileRef, 
	$outputFileRef, $evaluateRef, $modelsRef) = @_;
    
    # set default options
    $$wordsBeforeRef = $WINDOW_SIZE;
    $$wordsAfterRef = $WINDOW_SIZE;
    
    
    # get the options
    GetOptions('help', \&displayHelp,
	       'B=i', $wordsBeforeRef,
	       'A=i', $wordsAfterRef,
               'i=s',$sourceFileRef,
               'o=s', $outputFileRef,
	       'eval', $evaluateRef,
	       'model=s', $modelsRef);
    
    if(!defined $$sourceFileRef) {
	print STDERR "Please specify an input file, in SemCor format using the -i option\n";
	exit;
    }
    
    if(!defined $$outputFileRef) {
	print STDERR "Please specify an output file, using the -o option\n";
	exit;
    }
    
}




#=================================================================
# displayHelp
#
# prints out the help message, including usage 
#
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 02-19-2003    Rada            Initial version
# ========================================================================
sub displayHelp()
{
    print STDERR "Usage: senseLearner [-B N] [-A N] -i <input_file> -o <output-file> [-eval] -model <model> [-model <model>] \n\n";
    print STDERR "Minimally-supervised all words WSD\n\n";
    print STDERR "  -B N      the size of the context to the left of the target word,\n";
    print STDERR "            considered for feature extraction (default = 3)\n";
    print STDERR "  -A N      the size of the context to the right of the target word,\n";
    print STDERR "             considered for feature extraction (default = 3)\n";
    print STDERR "  -i <file> the input file, in SemCor format\n";
    print STDERR "  -o <file> the output file, to store the results\n";
    print STDERR "  -eval     run the evaluation, and output precision and coverage figures; \n";
    print STDERR "            the use of this option implies that correct word senses are\n";
    print STDERR "            available in the input file\n";
    print STDERR "  -model <model> the name of the model to be used for sense annotation; \n";
    print STDERR "            assumes the availability of a training instance base \n"; 
    print STDERR "            training.model and a subroutine that implements this model\n";

}



