#!/usr/bin/perl -w

# trainSenseLearnerModels.pl
#
# Script to create training instance bases, starting with a sense annotated
# file, for given learning models. For each sense tagged word in the input 
# file for which a learning model is provided, it creates a feature vector. 
# The class attribute for each such vector is the 
# target word with the corresponding sense.
#
# Copyright (c) 2004 Rada Mihalcea
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
#
#
# INPUT:    sense annotated file -- SemCor XML format 
# OUTPUT:   training data file
#
# CHANGE LOG
#
# Date          Author          Change
# =======================================================================
# 11-01-2004    Rada            add modules NNFeatures, VBFeatures, JJFeatures
#                               for additional flexibility
# 11-24-2003    Rada            initial version
# =======================================================================

use Getopt::Long;
use Word;
use Models;


# set default options and read command line options
$WINDOW_SIZE = 3;
$MODELS_DIR = "models";
if(scalar(@ARGV) == 0) {
    &displayHelp;
    exit;
}
&getCommandLineOptions(\$wordsBefore, \$wordsAfter, \$sourceFile, \@models);


# open the file with sense annotated data for reading
open ANNOTATEDFILE, "<$sourceFile";
(-r ANNOTATEDFILE) || die "Could not open source file $sourceFile. Please check path and make sure is valid\n";


my @fileHandles;
my $lenModels = scalar(@models);
# prepare the files for writing to store training feature vectors
for(my $m=0; $m < $lenModels; $m++) {
    open $fileHandles[$m], ">$MODELS_DIR/training.$models[$m]";
    (-w $fileHandles[$m]) || die "Could not open training file $MODELS_DIR/training.$models[$m]\n";
} 


# read one sentence at a time, and create training feature vectors
my (@sentence, $index);
while(my $line = <ANNOTATEDFILE>) {
    chomp $line;
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;

    # if start of sentence, clear up @sentence structure
    if($line =~ /^\<s snum=/) {
	undef @sentence;
	$index = 0;
    }    

    # if inside sentence, store Word-s
    if($line =~ /^\<wf/ ||
       $line =~ /^\<punc/) {
	$word = new Word($line);
	if(defined $word) {
	    $sentence[$index++] = $word;
	}
    }
    
    # if end of sentence, print out feature vectors
    if($line =~ /^\<\/s\>/) {
	
	# extract features for all models for this sentence
	for(my $m=0; $m < $lenModels; $m++) {
	    &{$models[$m]}(\@sentence, $fileHandles[$m], $wordsBefore, $wordsAfter);
	}
    }

}
    
# close all files 
close ANNOTATEDFILE;
for(my $m=0; $m < $lenModels; $m++) {
    close $fileHandles[$m];
} 


########################################################################
#
#  S U B R O U T I N E S
#
########################################################################


#==============================================================
# getCommandLineOptions
#
# Get the command lines options
# Check existence/validity of mandatory options
#
# CHANGE LOG
#
# Date          Author          Change
#=======================================================================
# 11-24-2003    Rada            initial version
#=======================================================================

sub getCommandLineOptions {
    my ($wordsBeforeRef, $wordsAfterRef, $sourceFileRef, $modelsRef) = @_;
    
    # set default options
    $$wordsBeforeRef = $WINDOW_SIZE;
    $$wordsAfterRef = $WINDOW_SIZE;

    # get the options
    GetOptions('help', \&displayHelp,
               'B=i', $wordsBeforeRef,
	       'A=i', $wordsAfterRef,
               'i=s', $sourceFileRef,
	       'model=s', $modelsRef);
    
    if(!defined $$sourceFileRef) {
	print STDERR "Please specify a source file, in SemCor XML format using the -i option\n";
	exit;
    }

    if(!defined @$modelsRef) {
	print STDERR "At least one learning model has to be specified, using the -model option\n";
	exit;
    }
    
}





#=================================================================
# displayHelp
#
# prints out the help message, including usage 
#
# CHANGE LOG
#
# Date          Author          Change
# ========================================================================
# 02-19-2003    Rada            Initial version
# ========================================================================

sub displayHelp()
{
    print STDERR "Usage: trainSenseLearnerModels [-B N] [-A N] -i <annotated_file> -model <model> [-model <model>] \n\n";
    print STDERR "This script creates one or more training instance bases starting with \n";
    print STDERR "a sense tagged corpus\n\n";
    print STDERR "  -B N      the size of the context to the left of the target word, \n";
    print STDERR "            considered for feature extraction (default = 3)\n";
    print STDERR "  -A N      the size of the context to the right of the target word, \n";
    print STDERR "            considered for feature extraction (default = 3)\n";
    print STDERR "  -i <file> the source file, including annotated data formatted using \n"; 
    print STDERR "            the SemCor XML format\n";
    print STDERR "  -model <model> the name of the model to be used for sense annotation; \n";
    print STDERR "            assumes the availability of a subroutine that implements \n"; 
    print STDERR "            this model; the trainining instance base is stored in a \n";
    print STDERR "            file named training.model; several models can be indicated \n"; 
    print STDERR "            at the same time\n";
}

